@@ -1,29 +1,30 @@
 use md6;
 use std::fmt;
 use serde_json;
+use std::ffi::OsStr;
 use std::borrow::Cow;
 use std::net::IpAddr;
 use serde::Serialize;
 use unicase::UniCase;
-use iron::mime::Mime;
 use std::sync::RwLock;
 use lazysort::SortedBy;
-use std::path::PathBuf;
 use cidr::{Cidr, IpCidr};
 use std::fs::{self, File};
 use std::default::Default;
 use rand::{Rng, thread_rng};
 use iron::modifiers::Header;
+use std::path::{PathBuf, Path};
 use iron::url::Url as GenericUrl;
-use mime_guess::guess_mime_type_opt;
+use mime_guess::get_mime_type_opt;
 use hyper_native_tls::NativeTlsServer;
 use std::collections::{BTreeMap, HashMap};
 use self::super::{LogLevel, Options, Error};
-use std::io::{self, SeekFrom, Write, Read, Seek};
 use std::process::{ExitStatus, Command, Child, Stdio};
 use rfsapi::{RawFsApiHeader, FilesetData, RawFileData};
 use rand::distributions::uniform::Uniform as UniformDistribution;
 use rand::distributions::Alphanumeric as AlphanumericDistribution;
+use iron::mime::{Mime, SubLevel as MimeSubLevel, TopLevel as MimeTopLevel};
+use std::io::{self, ErrorKind as IoErrorKind, SeekFrom, Write, Error as IoError, Read, Seek};
 use iron::{headers, status, method, mime, IronResult, Listening, Response, TypeMap, Request, Handler, Iron};
 use self::super::util::{WwwAuthenticate, DisplayThree, CommaList, Spaces, Dav, url_path, file_hash, is_symlink, encode_str, encode_file, file_length,
                         hash_string, html_response, file_binary, client_mobile, percent_decode, file_icon_suffix, is_actually_file, is_descendant_of,
@@ -33,46 +34,78 @@ use self::super::util::{WwwAuthenticate, DisplayThree, CommaList, Spaces, Dav, u
 
 
 macro_rules! log {
-    ($do_log:expr, $fmt:expr) => {
+    ($logcfg:expr, $fmt:expr) => {
         use time::now;
         use trivial_colours::{Reset as CReset, Colour as C};
 
-        if $do_log {
-            print!("{}[{}]{} ", C::Cyan, now().strftime("%F %T").unwrap(), CReset);
-            println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
-                     black = C::Black,
-                     red = C::Red,
-                     green = C::Green,
-                     yellow = C::Yellow,
-                     blue = C::Blue,
-                     magenta = C::Magenta,
-                     cyan = C::Cyan,
-                     white = C::White,
-                     reset = CReset);
+        if $logcfg.0 {
+            if $logcfg.1 {
+                print!("{}[{}]{} ", C::Cyan, now().strftime("%F %T").unwrap(), CReset);
+                println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
+                         black = C::Black,
+                         red = C::Red,
+                         green = C::Green,
+                         yellow = C::Yellow,
+                         blue = C::Blue,
+                         magenta = C::Magenta,
+                         cyan = C::Cyan,
+                         white = C::White,
+                         reset = CReset);
+            } else {
+                print!("[{}] ", now().strftime("%F %T").unwrap());
+                println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
+                         black = "",
+                         red = "",
+                         green = "",
+                         yellow = "",
+                         blue = "",
+                         magenta = "",
+                         cyan = "",
+                         white = "",
+                         reset = "");
+            }
         }
     };
-    ($do_log:expr, $fmt:expr, $($arg:tt)*) => {
+    ($logcfg:expr, $fmt:expr, $($arg:tt)*) => {
         use time::now;
         use trivial_colours::{Reset as CReset, Colour as C};
 
-        if $do_log {
-            print!("{}[{}]{} ", C::Cyan, now().strftime("%F %T").unwrap(), CReset);
-            println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
-                     $($arg)*,
-                     black = C::Black,
-                     red = C::Red,
-                     green = C::Green,
-                     yellow = C::Yellow,
-                     blue = C::Blue,
-                     magenta = C::Magenta,
-                     cyan = C::Cyan,
-                     white = C::White,
-                     reset = CReset);
+        if $logcfg.0 {
+            if $logcfg.1 {
+                print!("{}[{}]{} ", C::Cyan, now().strftime("%F %T").unwrap(), CReset);
+                println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
+                         $($arg)*,
+                         black = C::Black,
+                         red = C::Red,
+                         green = C::Green,
+                         yellow = C::Yellow,
+                         blue = C::Blue,
+                         magenta = C::Magenta,
+                         cyan = C::Cyan,
+                         white = C::White,
+                         reset = CReset);
+            } else {
+                print!("[{}] ", now().strftime("%F %T").unwrap());
+                println!(concat!($fmt, "{black:.0}{red:.0}{green:.0}{yellow:.0}{blue:.0}{magenta:.0}{cyan:.0}{white:.0}{reset:.0}"),
+                         $($arg)*,
+                         black = "",
+                         red = "",
+                         green = "",
+                         yellow = "",
+                         blue = "",
+                         magenta = "",
+                         cyan = "",
+                         white = "",
+                         reset = "");
+            }
         }
     };
 }
 
 mod webdav;
+mod bandwidth;
+
+pub use self::bandwidth::{LimitBandwidthMiddleware, SimpleChain};
 
 
 // TODO: ideally this String here would be Encoding instead but hyper is bad
@@ -83,13 +116,16 @@ pub struct HttpHandler {
     pub follow_symlinks: bool,
     pub sandbox_symlinks: bool,
     pub check_indices: bool,
-    pub log: bool,
+    pub strip_extensions: bool,
+    /// (at all, log_colour)
+    pub log: (bool, bool),
     pub webdav: bool,
     pub global_auth_data: Option<(String, Option<String>)>,
     pub path_auth_data: BTreeMap<String, Option<(String, Option<String>)>>,
     pub writes_temp_dir: Option<(String, PathBuf)>,
     pub encoded_temp_dir: Option<(String, PathBuf)>,
     pub proxies: BTreeMap<IpCidr, String>,
+    pub mime_type_overrides: BTreeMap<String, Mime>,
     cache_gen: RwLock<CacheT<Vec<u8>>>,
     cache_fs: RwLock<CacheT<(PathBuf, bool)>>,
 }
@@ -118,7 +154,8 @@ impl HttpHandler {
             follow_symlinks: opts.follow_symlinks,
             sandbox_symlinks: opts.sandbox_symlinks,
             check_indices: opts.check_indices,
-            log: opts.loglevel < LogLevel::NoServeStatus,
+            strip_extensions: opts.strip_extensions,
+            log: (opts.loglevel < LogLevel::NoServeStatus, opts.log_colour),
             webdav: opts.webdav,
             global_auth_data: global_auth_data,
             path_auth_data: path_auth_data,
@@ -127,13 +164,16 @@ impl HttpHandler {
             cache_gen: Default::default(),
             cache_fs: Default::default(),
             proxies: opts.proxies.clone(),
+            mime_type_overrides: opts.mime_type_overrides.clone(),
         }
     }
 
-    pub fn clean_temp_dirs(temp_dir: &(String, PathBuf), loglevel: LogLevel) {
+    pub fn clean_temp_dirs(temp_dir: &(String, PathBuf), loglevel: LogLevel, log_colour: bool) {
         for (temp_name, temp_dir) in ["writes", "encoded", "tls"].iter().flat_map(|tn| HttpHandler::temp_subdir(temp_dir, true, tn)) {
             if temp_dir.exists() && fs::remove_dir_all(&temp_dir).is_ok() {
-                log!(loglevel < LogLevel::NoServeStatus, "Deleted temp dir {magenta}{}{reset}", temp_name);
+                log!((loglevel < LogLevel::NoServeStatus, log_colour),
+                     "Deleted temp dir {magenta}{}{reset}",
+                     temp_name);
             }
         }
     }
@@ -287,12 +327,18 @@ impl HttpHandler {
     }
 
     fn handle_get(&self, req: &mut Request) -> IronResult<Response> {
-        let (req_p, symlink, url_err) = self.parse_requested_path(req);
+        let (mut req_p, symlink, url_err) = self.parse_requested_path(req);
 
         if url_err {
             return self.handle_invalid_url(req, "<p>Percent-encoding decoded to invalid UTF-8.</p>");
         }
 
+        if !req_p.exists() && req_p.extension().is_none() && self.strip_extensions {
+            if let Some(rp) = INDEX_EXTENSIONS.iter().map(|ext| req_p.with_extension(ext)).find(|rp| rp.exists()) {
+                req_p = rp;
+            }
+        }
+
         if !req_p.exists() || (symlink && !self.follow_symlinks) ||
            (symlink && self.follow_symlinks && self.sandbox_symlinks && !is_descendant_of(&req_p, &self.hosted_directory.1)) {
             return self.handle_nonexistent(req, req_p);
@@ -396,11 +442,7 @@ impl HttpHandler {
     }
 
     fn handle_get_file_closed_range(&self, req: &mut Request, req_p: PathBuf, from: u64, to: u64) -> IronResult<Response> {
-        let mime_type = guess_mime_type_opt(&req_p).unwrap_or_else(|| if file_binary(&req_p) {
-            "application/octet-stream".parse().unwrap()
-        } else {
-            "text/plain".parse().unwrap()
-        });
+        let mime_type = self.guess_mime_type(&req_p);
         log!(self.log,
              "{} was served byte range {}-{} of file {magenta}{}{reset} as {blue}{}{reset}",
              self.remote_addresses(&req),
@@ -427,11 +469,7 @@ impl HttpHandler {
     }
 
     fn handle_get_file_right_opened_range(&self, req: &mut Request, req_p: PathBuf, from: u64) -> IronResult<Response> {
-        let mime_type = guess_mime_type_opt(&req_p).unwrap_or_else(|| if file_binary(&req_p) {
-            "application/octet-stream".parse().unwrap()
-        } else {
-            "text/plain".parse().unwrap()
-        });
+        let mime_type = self.guess_mime_type(&req_p);
         log!(self.log,
              "{} was served file {magenta}{}{reset} from byte {} as {blue}{}{reset}",
              self.remote_addresses(&req),
@@ -444,11 +482,7 @@ impl HttpHandler {
     }
 
     fn handle_get_file_left_opened_range(&self, req: &mut Request, req_p: PathBuf, from: u64) -> IronResult<Response> {
-        let mime_type = guess_mime_type_opt(&req_p).unwrap_or_else(|| if file_binary(&req_p) {
-            "application/octet-stream".parse().unwrap()
-        } else {
-            "text/plain".parse().unwrap()
-        });
+        let mime_type = self.guess_mime_type(&req_p);
         log!(self.log,
              "{} was served last {} bytes of file {magenta}{}{reset} as {blue}{}{reset}",
              self.remote_addresses(&req),
@@ -491,11 +525,7 @@ impl HttpHandler {
     }
 
     fn handle_get_file_empty_range(&self, req: &mut Request, req_p: PathBuf, from: u64, to: u64) -> IronResult<Response> {
-        let mime_type = guess_mime_type_opt(&req_p).unwrap_or_else(|| if file_binary(&req_p) {
-            "application/octet-stream".parse().unwrap()
-        } else {
-            "text/plain".parse().unwrap()
-        });
+        let mime_type = self.guess_mime_type(&req_p);
         log!(self.log,
              "{} was served an empty range from file {magenta}{}{reset} as {blue}{}{reset}",
              self.remote_addresses(&req),
@@ -514,11 +544,7 @@ impl HttpHandler {
     }
 
     fn handle_get_file(&self, req: &mut Request, req_p: PathBuf) -> IronResult<Response> {
-        let mime_type = guess_mime_type_opt(&req_p).unwrap_or_else(|| if file_binary(&req_p) {
-            "application/octet-stream".parse().unwrap()
-        } else {
-            "text/plain".parse().unwrap()
-        });
+        let mime_type = self.guess_mime_type(&req_p);
         log!(self.log,
              "{} was served file {magenta}{}{reset} as {blue}{}{reset}",
              self.remote_addresses(&req),
@@ -531,11 +557,15 @@ impl HttpHandler {
            req_p.extension().and_then(|s| s.to_str()).map(|s| !BLACKLISTED_ENCODING_EXTENSIONS.contains(&UniCase::new(s))).unwrap_or(true) {
             self.handle_get_file_encoded(req, req_p, mime_type)
         } else {
+            let file = match File::open(&req_p) {
+                Ok(file) => file,
+                Err(err) => return self.handle_requested_entity_unopenable(req, err, "file"),
+            };
             Ok(Response::with((status::Ok,
                                (Header(headers::Server(USER_AGENT.to_string())),
                                 Header(headers::LastModified(headers::HttpDate(file_time_modified(&metadata)))),
                                 Header(headers::AcceptRanges(vec![headers::RangeUnit::Bytes]))),
-                               req_p.as_path(),
+                               file,
                                Header(headers::ContentLength(file_length(&metadata, &req_p))),
                                mime_type)))
         }
@@ -633,7 +663,8 @@ impl HttpHandler {
         self.handle_raw_fs_api_response(status::Ok,
                                         &FilesetData {
                                             writes_supported: self.writes_temp_dir.is_some(),
-                                            is_root: req.url.as_ref().path_segments().unwrap().count() == 1,
+                                            is_root: req.url.as_ref().path_segments().unwrap().count() + !req.url.as_ref().as_str().ends_with('/') as usize ==
+                                                     1,
                                             is_file: false,
                                             files: req_p.read_dir()
                                                 .expect("Failed to read requested directory")
@@ -709,7 +740,7 @@ impl HttpHandler {
         //     https://cloud.githubusercontent.com/assets/6709544/21442017/9eb20d64-c89b-11e6-8c7b-888b5f70a403.png
         //   - With following slash:
         //     https://cloud.githubusercontent.com/assets/6709544/21442028/a50918c4-c89b-11e6-8936-c29896947f6a.png
-        Ok(Response::with((status::MovedPermanently, Header(headers::Server(USER_AGENT.to_string())), Header(headers::Location(new_url)))))
+        Ok(Response::with((status::SeeOther, Header(headers::Server(USER_AGENT.to_string())), Header(headers::Location(new_url)))))
     }
 
     fn handle_get_mobile_dir_listing(&self, req: &mut Request, req_p: PathBuf) -> IronResult<Response> {
@@ -732,7 +763,7 @@ impl HttpHandler {
                     file_time_modified_p(req_p.parent().expect("Failed to get requested directory's parent directory"))
                         .strftime("%F %T")
                         .unwrap(),
-                    up_path = slash_idx.map(|i| &rel_noslash[0..i]).unwrap_or(""),
+                    up_path = slash_idx.map(|i| &rel_noslash[0..i]).unwrap_or("").replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"),
                     up_path_slash = if slash_idx.is_some() { "/" } else { "" })
         };
         let list_s = req_p.read_dir()
@@ -784,8 +815,8 @@ impl HttpHandler {
                         } else {
                             DisplayThree("", String::new(), "")
                         },
-                        path = format!("/{}", relpath).replace("//", "/").replace('%', "%25").replace('#', "%23"),
-                        fname = fname.replace('%', "%25").replace('#', "%23"))
+                        path = format!("/{}", relpath).replace("//", "/").replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"),
+                        fname = fname.replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"))
             });
 
         self.handle_generated_response_encoding(req,
@@ -835,12 +866,16 @@ impl HttpHandler {
                          <td><a href=\"/{up_path}{up_path_slash}\">&nbsp;</a></td> \
                          <td><a href=\"/{up_path}{up_path_slash}\">&nbsp;</a></td></tr>",
                     file_time_modified_p(req_p.parent().expect("Failed to get requested directory's parent directory")).strftime("%F %T").unwrap(),
-                    up_path = slash_idx.map(|i| &rel_noslash[0..i]).unwrap_or(""),
+                    up_path = slash_idx.map(|i| &rel_noslash[0..i]).unwrap_or("").replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"),
                     up_path_slash = if slash_idx.is_some() { "/" } else { "" })
         };
-        let list_s = req_p.read_dir()
-            .expect("Failed to read requested directory")
-            .map(|p| p.expect("Failed to iterate over requested directory"))
+
+
+        let rd = match req_p.read_dir() {
+            Ok(rd) => rd,
+            Err(err) => return self.handle_requested_entity_unopenable(req, err, "directory"),
+        };
+        let list_s = rd.map(|p| p.expect("Failed to iterate over requested directory"))
             .filter(|f| {
                 let fp = f.path();
                 let mut symlink = false;
@@ -895,8 +930,8 @@ impl HttpHandler {
                         } else {
                             DisplayThree("", "", "")
                         },
-                        path = format!("/{}", relpath).replace("//", "/").replace('%', "%25").replace('#', "%23"),
-                        fname = fname.replace('%', "%25").replace('#', "%23"))
+                        path = format!("/{}", relpath).replace("//", "/").replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"),
+                        fname = fname.replace('%', "%25").replace('#', "%23").replace('[', "%5B").replace(']', "%5D"))
             });
 
         self.handle_generated_response_encoding(req,
@@ -1194,6 +1229,18 @@ impl HttpHandler {
         Ok(Response::with((st, Header(headers::Server(USER_AGENT.to_string())), "text/html;charset=utf-8".parse::<mime::Mime>().unwrap(), resp)))
     }
 
+    fn handle_requested_entity_unopenable(&self, req: &mut Request, e: IoError, entity_type: &str) -> IronResult<Response> {
+        if e.kind() == IoErrorKind::PermissionDenied {
+            self.handle_generated_response_encoding(req,
+                                                    status::Forbidden,
+                                                    html_response(ERROR_HTML, &["403 Forbidden", &format!("Can't access {}.", url_path(&req.url)), ""]))
+        } else {
+            // The ops that get here (File::open(), fs::read_dir()) can't return any other errors by the time they're run
+            // (and even if it could, there isn't much we can do about them)
+            panic!("Failed to read requested {}: {:?}", entity_type, e)
+        }
+    }
+
     fn handle_raw_fs_api_response<R: Serialize>(&self, st: status::Status, resp: &R) -> IronResult<Response> {
         Ok(Response::with((st,
                            Header(headers::Server(USER_AGENT.to_string())),
@@ -1255,8 +1302,22 @@ impl HttpHandler {
         AddressWriter {
             request: req,
             proxies: &self.proxies,
+            log: self.log,
         }
     }
+
+    fn guess_mime_type(&self, req_p: &Path) -> Mime {
+        // Based on mime_guess::guess_mime_type_opt(); that one does to_str() instead of to_string_lossy()
+        let ext = req_p.extension().map(OsStr::to_string_lossy).unwrap_or("".into());
+
+        (self.mime_type_overrides.get(&*ext).cloned())
+            .or_else(|| get_mime_type_opt(&*ext))
+            .unwrap_or_else(|| if file_binary(req_p) {
+                Mime(MimeTopLevel::Application, MimeSubLevel::OctetStream, Default::default()) // "application/octet-stream"
+            } else {
+                Mime(MimeTopLevel::Text, MimeSubLevel::Plain, Default::default()) // "text/plain"
+            })
+    }
 }
 
 impl Clone for HttpHandler {
@@ -1266,6 +1327,7 @@ impl Clone for HttpHandler {
             follow_symlinks: self.follow_symlinks,
             sandbox_symlinks: self.sandbox_symlinks,
             check_indices: self.check_indices,
+            strip_extensions: self.strip_extensions,
             log: self.log,
             webdav: self.webdav,
             global_auth_data: self.global_auth_data.clone(),
@@ -1273,6 +1335,7 @@ impl Clone for HttpHandler {
             writes_temp_dir: self.writes_temp_dir.clone(),
             encoded_temp_dir: self.encoded_temp_dir.clone(),
             proxies: self.proxies.clone(),
+            mime_type_overrides: self.mime_type_overrides.clone(),
             cache_gen: Default::default(),
             cache_fs: Default::default(),
         }
@@ -1283,19 +1346,29 @@ impl Clone for HttpHandler {
 pub struct AddressWriter<'r, 'p, 'ra, 'rb: 'ra> {
     pub request: &'r Request<'ra, 'rb>,
     pub proxies: &'p BTreeMap<IpCidr, String>,
+    /// (at all, log_colour)
+    pub log: (bool, bool),
 }
 
 impl<'r, 'p, 'ra, 'rb: 'ra> fmt::Display for AddressWriter<'r, 'p, 'ra, 'rb> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         use trivial_colours::{Reset as CReset, Colour as C};
 
-        write!(f, "{green}{}{reset}", self.request.remote_addr, green = C::Green, reset = CReset)?;
+        if self.log.1 {
+            write!(f, "{green}{}{reset}", self.request.remote_addr, green = C::Green, reset = CReset)?;
+        } else {
+            write!(f, "{}", self.request.remote_addr)?;
+        }
 
         for (network, header) in self.proxies {
             if network.contains(&self.request.remote_addr.ip()) {
                 if let Some(saddrs) = self.request.headers.get_raw(header) {
                     for saddr in saddrs {
-                        write!(f, " for {green}{}{reset}", String::from_utf8_lossy(saddr), green = C::Green, reset = CReset)?;
+                        if self.log.1 {
+                            write!(f, " for {green}{}{reset}", String::from_utf8_lossy(saddr), green = C::Green, reset = CReset)?;
+                        } else {
+                            write!(f, " for {}", String::from_utf8_lossy(saddr))?;
+                        }
                     }
                 }
             }
